/*-
 * Copyright (c) 1999 Assar Westerlund
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD: src/share/examples/kld/syscall/module/syscall.c,v 1.3.8.1 2004/09/28 18:27:28 pjd Exp $
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/proc.h>
#include <sys/module.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/sysent.h>
#include <sys/kernel.h>
#include <sys/kthread.h>
#include <sys/systm.h>
#include <sys/sched.h>

// -----
volatile static int	tdone = 0;

// -----

static inline uint64_t spin(uint64_t start, uint64_t limit);
static inline uint64_t
spin(uint64_t start, uint64_t limit)
{
        uint64_t        delta;
        do {
                delta = rdtsc() - start;
        } while (delta < limit);
        return(delta);  
}

// -----

static void helper(void *arg);

static void
helper(void *arg)
{
        uint64_t        tstart = 0;
        uint64_t        tstop = 0;
        uint64_t        start = 0;
        uint32_t        count = 10000000;
        uint64_t        avg = 0;
        uint64_t        avg_l = 0;
        uint32_t        count_l=0;
        uint64_t        tavg = 0;
        uint64_t        delta = 0;
        uint16_t        limit = 100;

printf("helper running\n");


	// bind/pin
	printf("pinning\n");
	mtx_lock_spin(&sched_lock);
	sched_bind(curthread,2);
	sched_pin();
	mtx_unlock_spin(&sched_lock);

// disable_intr();
	
	tstart = rdtsc();
        for(uint32_t i=0; i<count; i++) {
                start = rdtsc();
                delta = spin(start, limit);
                if (delta > (limit+300)) {
                        avg_l += delta;
                        count_l ++;
                } else {
                        avg += delta;
                }
                tavg += rdtsc() - start;
        }
	tstop = rdtsc();
// enable_intr();

	printf("cycles = %10ju - cycles per abnormality = %10ju\n",
		tstop-tstart, (tstop-tstart)/count_l);

        printf("desired_avg: %6ju - long_avg: %6ju - num_long: %10u - total_avg: %ju\n", avg/(count-count_l), avg_l/count_l, count_l, tavg/count);

	// unpin/unbind
	mtx_lock_spin(&sched_lock);
	sched_unpin();
	sched_unbind(curthread);
	mtx_unlock_spin(&sched_lock);
	printf("unpinned\n");

	tdone = 1;

	kthread_exit(0);
}

// -----

static int
kmypin (struct thread *td, void *arg)
{
	tdone = 0;
	int	bob;

uprintf("starting helper\n");
	kthread_create(helper, NULL, NULL, 0, 0, "helper%d", 0);
uprintf("waiting for helper to finish\n");

	while (!tdone) {
		tsleep(&bob, 0, "xxx", 1);
	}

	// give the other thread a chance to exit
	tsleep(&bob, 0, "xxx", 100);
uprintf("helper finished\n");

	return (0);
}

/*
 * The `sysent' for the new syscall
 */

static struct sysent kmypin_sysent = {
	0,			/* sy_narg */
	kmypin			/* sy_call */
};

/*
 * The offset in sysent where the syscall is allocated.
 */

static int offset = NO_SYSCALL;

/*
 * The function called at load/unload.
 */

static int
load (struct module *module, int cmd, void *arg)
{
	int error = 0;

	switch (cmd) {
	case MOD_LOAD :
		printf ("syscall loaded at %d\n", offset);
		break;
	case MOD_UNLOAD :
		printf ("syscall unloaded from %d\n", offset);
		break;
	default :
		error = EOPNOTSUPP;
		break;
	}
	return error;
}

SYSCALL_MODULE(kpinsys, &offset, &kmypin_sysent, load, NULL);
