/*-
 * Copyright (c) 1999 Assar Westerlund
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD: src/share/examples/kld/syscall/test/call.c,v 1.3 1999/08/28 00:19:24 peter Exp $
 */

#include <stdio.h>
#include <sys/syscall.h>
#include <sys/types.h>
#include <sys/module.h>
#include <stdlib.h>
#include <unistd.h>
#include <machine/cpufunc.h>
#include <pthread.h>
#include <pthread_np.h>

// -----

volatile	int	trunning = 0;
volatile	int	tstart = 0;
volatile	int	tdone = 0;

// -----

static inline uint64_t spin(uint64_t start, uint64_t limit);
static inline uint64_t
spin(uint64_t start, uint64_t limit)
{
        uint64_t        delta;
        do {
                delta = rdtsc() - start;
        } while (delta < limit);
        return(delta);  
}


// -----

void* helper(void *arg);

void*
helper(void *arg)
{
	char *endptr;
	int syscall_num;
	struct module_stat stat;
	uint64_t	fstart = 0;
	uint64_t	fstop = 0;
	uint64_t	start = 0;
	uint64_t	stop = 0;
	uint32_t	count = 10000000;
        uint64_t        avg = 0;
        uint64_t        avg_l = 0;
        uint32_t        count_l=0;
        uint64_t        tavg = 0;
        uint64_t        delta = 0;
	uint16_t	limit = 100;

	// tell main that we are running
	trunning = 1;

	// wait while main checks if we are system scope
	while (!tstart);

	// find binding and pinning system call module
	stat.version = sizeof(stat);
	modstat(modfind("pinsys"), &stat);
	syscall_num = stat.data.intval;

	// toggle into bound and pinned mode
	syscall(syscall_num);

	// try waiting for "limit" cycles
	// if we actually wait "limit + 300" cycles we've context swapped
	// there should be zero context swaps if we are bound and pinned
	fstart = rdtsc();
        for(uint32_t i=0; i<count; i++) {
                start = rdtsc();
                delta = spin(start, limit);
                if (delta > (limit+300)) {
                        avg_l += delta;
                        count_l ++;
                } else {
                        avg += delta;
                }
                tavg += rdtsc() - start;
        }
	fstop = rdtsc();

	printf("cycles = %10ju - cycles per abnormality = %10ju\n", 
		fstop-fstart, (fstop-fstart)/count_l);

        printf("desired_avg: %6ju - long_avg: %6ju - num_long: %10u - total_avg: %ju\n", avg/(count-count_l), avg_l/count_l, count_l, tavg/count);

	// toggle out of bound and pinned mode
	syscall(syscall_num);

	// tell main that we are done
	tdone = 1;

	pthread_exit(NULL);
}

// -----

int
main(int argc, char **argv)
{
	pthread_t	pt;
	pthread_attr_t	pa;
	pthread_attr_t	pan;
	int		scope;

	// create a new attr and set system scope on it
	pthread_attr_init(&pa);
	scope = PTHREAD_SCOPE_SYSTEM;
	pthread_attr_setscope(&pa, scope);

	// create a new system scope thread
	pthread_create(&pt, &pa, helper, NULL);

	// wait for thread to start
	while (!trunning);


	// read new thread's scope
	pthread_attr_init(&pan);
	pthread_attr_get_np(pt, &pan);
	scope = ~0;
	pthread_attr_getscope(&pan, &scope);
	printf("scope = %d\n", scope);

	// let other thread start
	tstart = 1;

	// wait for it to finish - thread should be bound
	// so on SMP/CMP system spin waiting is fine
	while (!tdone);

	pthread_join(pt, NULL);
}
